#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import sys
import time

sys.path.append(os.path.join(os.path.dirname(__file__), "../src/"))

import asyncio
from agefreighter import AgeFreighter

import networkx as nx
import pandas as pd
import nest_asyncio
from psycopg.rows import namedtuple_row

nest_asyncio.apply()


async def is_graph_created(
    af: AgeFreighter,
    graph_name: str = "",
    vertex_labels: list = [],
    vertex_counts: list = [],
    edge_type: str = "",
    edge_count: int = 0,
) -> None:
    """
    Check the number of vertices and edges in the graph
    """
    result = True
    if graph_name.lower() != graph_name:
        graph_name = f'"{graph_name}"'
    async with af.pool.connection() as conn:
        async with conn.cursor(row_factory=namedtuple_row) as cur:
            for idx, (v_label, v_count) in enumerate(zip(vertex_labels, vertex_counts)):
                await cur.execute(f'SELECT COUNT(*) FROM {graph_name}."{v_label}";')
                cnt_result = await cur.fetchone()
                result &= cnt_result.count == v_count
            await cur.execute(f'SELECT COUNT(*) FROM {graph_name}."{edge_type}";')
            cnt_result = await cur.fetchone()
            result &= cnt_result.count == edge_count
    return result


async def show_test_result(
    func_name: str = "",
    execution_time: float = 0.0,
    chunk_size: int = 0,
    direct_loading: bool = False,
    use_copy: bool = False,
) -> None:
    print(
        f"{func_name} : time, {execution_time:.2f}, chunk_size: {chunk_size}, direct_loading: {direct_loading}, use_copy: {use_copy}"
    )


async def test_loadFromSingleCSV(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
) -> None:
    """
    Test for loadFromSingleCSV()
    file downloaded from https://www.kaggle.com/datasets/darinhawley/imdb-films-by-actor-for-10k-actors
    actorfilms.csv: Actor,ActorID,Film,Year,Votes,Rating,FilmID
    # of actors: 9,623, # of films: 44,456, # of edges: 191,873
    """
    graph_name = "ActorFilmsSingleCSV"
    start_v_label = "Actor"
    edge_type = "ACTED_IN"
    end_v_label = "Film"
    start_time = time.time()
    await af.loadFromSingleCSV(
        graph_name=graph_name,
        csv="../data/actorfilms.csv",
        start_v_label=start_v_label,
        start_id="ActorID",
        start_props=["Actor"],
        edge_type=edge_type,
        end_v_label=end_v_label,
        end_id="FilmID",
        end_props=["Film", "Year", "Votes", "Rating"],
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=[start_v_label, end_v_label],
        vertex_counts=[9623, 44456],
        edge_type=edge_type,
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def test_loadFromCSVs(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
) -> None:
    """
    Test for loadFromCSVs()
    cities.csv: id,name,state_id,state_code,country_id,country_code,latitude,longitude
    countries.csv: id,name,iso3,iso2,numeric_code,phone_code,capital,currency,currency_symbol,tld,native,region,subregion,latitude,longitude,emoji,emojiU
    edges.csv: start_id,start_vertex_type,end_id,end_vertex_type
    # of countries: 53, # of cities: 72,485, # of edges: 72,485
    """
    graph_name = "cities_countries_from_csvs"
    vertex_labels = ["Country", "City"]
    edge_types = ["has_city"]
    start_time = time.time()
    await af.loadFromCSVs(
        graph_name=graph_name,
        vertex_csvs=["../data/countries.csv", "../data/cities.csv"],
        vertex_labels=vertex_labels,
        edge_csvs=["../data/edges.csv"],
        edge_types=edge_types,
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=vertex_labels,
        vertex_counts=[53, 72485],
        edge_type=edge_types[0],
        edge_count=72485,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def test_loadFrom2CSVs(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
) -> None:
    """
    Test for loadFromCSVs()
    start and end vertices are in the same csv file
    """
    graph_name = "war_btw_countries_from_2csvs"
    vertex_labels = ["Country"]
    edge_types = ["FIGHT_WITH"]
    start_time = time.time()
    await af.loadFromCSVs(
        graph_name=graph_name,
        vertex_csvs=["../data/countries.csv"],
        vertex_labels=vertex_labels,
        edge_csvs=["../data/fight_with.csv"],
        edge_types=edge_types,
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=vertex_labels,
        vertex_counts=[53],
        edge_type=edge_types[0],
        edge_count=3,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def test_loadFromNetworkx(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
) -> None:
    """
    Test for loadFromNetworkx()
    create networkx graph from actorfilms.csv
    file downloaded from https://www.kaggle.com/datasets/darinhawley/imdb-films-by-actor-for-10k-actors
    after creating networkx graph, load it to the database
    """
    df = pd.read_csv("../data/actorfilms.csv")
    G = nx.DiGraph()

    for name, group in df.groupby("ActorID"):
        for idx, row in group.iterrows():
            G.add_node(row["ActorID"], label="Actor", name=row["Actor"])
            G.add_node(
                row["FilmID"],
                label="Film",
                name=row["Film"],
                year=row["Year"],
                votes=row["Votes"],
                rating=row["Rating"],
            )
            G.add_edge(row["ActorID"], row["FilmID"], label="ACTED_IN")

    graph_name = "ActorFilmsFromNetworkx"
    id_map = {
        "Actor": "ActorID",
        "Film": "FilmID",
    }
    start_time = time.time()
    await af.loadFromNetworkx(
        graph_name=graph_name,
        networkx_graph=G,
        id_map=id_map,
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=["Actor", "Film"],
        vertex_counts=[9623, 44456],
        edge_type="ACTED_IN",
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def test_loadFromNeo4j(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
    init_neo4j: bool = False,
) -> None:
    """
    Test for loadFromNeo4j()
    create networkx graph from actorfilms.csv
    after creating networkx graph, load it to a graph
    """
    try:
        n4j_uri = os.environ["NEO4J_URI"]
        n4j_user = os.environ["NEO4J_USER"]
        n4j_password = os.environ["NEO4J_PASSWORD"]
    except KeyError:
        print(
            "Please set the environment variables NEO4J_URI, NEO4J_USER, NEO4J_PASSWORD"
        )
        return

    # prepare test data for neo4j
    if init_neo4j:
        await loadTestDataToNeo4j(n4j_uri, n4j_user, n4j_password)

    start_time = time.time()
    graph_name = "Actor_FilmsFromNeo4j"
    id_map = {"Actor": "ActorID", "Film": "FilmID"}
    await af.loadFromNeo4j(
        uri=n4j_uri,
        user=n4j_user,
        password=n4j_password,
        neo4j_database="neo4j",
        graph_name=graph_name,
        id_map=id_map,
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=id_map.keys(),
        vertex_counts=[9623, 44456],
        edge_type="ACTED_IN",
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def loadTestDataToNeo4j(
    n4j_uri: str = "",
    n4j_user: str = "",
    n4j_password: str = "",
) -> None:
    """
    Load test data to Neo4j
    file downloaded from https://www.kaggle.com/datasets/darinhawley/imdb-films-by-actor-for-10k-actors
    """
    from neo4j import AsyncGraphDatabase

    batch_size = 1000
    df = pd.read_csv("../data/actorfilms.csv")
    uniq_actors = df[["ActorID", "Actor"]].drop_duplicates()
    uniq_films = df[["FilmID", "Film", "Year", "Votes", "Rating"]].drop_duplicates()

    async with AsyncGraphDatabase.driver(
        n4j_uri, auth=(n4j_user, n4j_password)
    ) as driver:
        async with driver.session() as session:
            # clear the database
            await session.run("MATCH (a)-[r]->() DELETE a, r")
            await session.run("MATCH (a) DELETE a")
            await session.run("DROP INDEX actor_index_id IF EXISTS")
            await session.run("DROP INDEX film_index_id IF EXISTS")
            await session.run(
                "CREATE INDEX actor_index_id FOR (n:Actor) ON (n.ActorID)"
            )
            await session.run("CREATE INDEX film_index_id FOR (n:Film) ON (n.FilmID)")
            # create actor nodes
            for idx in range(0, len(uniq_actors), batch_size):
                actors = [
                    {"Actor": actor, "ActorID": actorid}
                    for i, (actor, actorid) in enumerate(
                        zip(
                            uniq_actors["Actor"][idx : idx + batch_size].tolist(),
                            uniq_actors["ActorID"][idx : idx + batch_size].tolist(),
                        )
                    )
                ]
                await session.run(
                    """UNWIND $actors AS row
                    CREATE (a:Actor)
                    SET a += row""",
                    actors=actors,
                )
            # create film nodes
            for idx in range(0, len(uniq_films), batch_size):
                films = [
                    {
                        "Film": film,
                        "FilmID": filmid,
                        "Year": year,
                        "Votes": votes,
                        "Rating": rating,
                    }
                    for i, (film, filmid, year, votes, rating) in enumerate(
                        zip(
                            uniq_films["Film"][idx : idx + batch_size].tolist(),
                            uniq_films["FilmID"][idx : idx + batch_size].tolist(),
                            uniq_films["Year"][idx : idx + batch_size].tolist(),
                            uniq_films["Votes"][idx : idx + batch_size].tolist(),
                            uniq_films["Rating"][idx : idx + batch_size].tolist(),
                        )
                    )
                ]
                await session.run(
                    """UNWIND $films AS row
                    CREATE (f:Film)
                    SET f += row""",
                    films=films,
                )
            # create edges
            for idx in range(0, len(df), batch_size):
                acted_ins = [
                    {"from": actorid, "to": filmid}
                    for i, (actorid, filmid) in enumerate(
                        zip(
                            df["ActorID"][idx : idx + batch_size].tolist(),
                            df["FilmID"][idx : idx + batch_size].tolist(),
                        )
                    )
                ]
                await session.run(
                    """UNWIND $acted_ins AS row
                    MATCH (from:Actor {ActorID: row.from})
                    MATCH (to:Film {FilmID: row.to})
                    CREATE (from)-[r:ACTED_IN]->(to)
                    SET r += row""",
                    acted_ins=acted_ins,
                )


async def test_loadFromPGSQL(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
    init_pgsql: bool = False,
) -> None:
    """
    Test for loadFromPGSQL()
    create tables from actorfilms.csv
    after creating table, load it to a graph
    """
    try:
        src_connection_string = os.environ["SRC_PG_CONNECTION_STRING"]
    except KeyError:
        print("Please set the environment variables SRC_PG_CONNECTION_STRING")
        return

    src_tables = {"from_nodes": "Actor", "to_nodes": "Film", "edges": "ACTED_IN"}

    if init_pgsql:
        # prepare test data for pgsql
        await loadTestDataToPGSQL(
            con_string=src_connection_string,
            src_tables=src_tables,
            src_csv="../data/actorfilms.csv",
        )

    start_time = time.time()
    graph_name = "actorfilmsFromPGSQL"
    id_map = {
        "Actor": "actorid",
        "Film": "filmid",
    }
    await af.loadFromPGSQL(
        src_con_string=src_connection_string,
        src_tables=src_tables,
        graph_name=graph_name,
        # values are culumn name with small caps
        id_map=id_map,
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=id_map.keys(),
        vertex_counts=[9623, 44456],
        edge_type="ACTED_IN",
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def loadTestDataToPGSQL(
    con_string: str = "",
    src_tables: dict = {},
    src_csv: str = "",
) -> None:
    """
    Load test data to PGSQL
    file downloaded from https://www.kaggle.com/datasets/darinhawley/imdb-films-by-actor-for-10k-actors
    """
    import psycopg as pg

    df = pd.read_csv(src_csv)

    datum = [None, None, None]
    types = [None, None, None]

    datum[0] = df[["ActorID", "Actor"]].drop_duplicates()
    datum[0].insert(0, "serial", range(1, len(datum[0]) + 1))
    types[0] = ["SERIAL", "TEXT", "TEXT"]

    datum[1] = df[["FilmID", "Film", "Year", "Votes", "Rating"]].drop_duplicates()
    datum[1].insert(0, "serial", range(1, len(datum[1]) + 1))
    types[1] = ["SERIAL", "TEXT", "TEXT", "INT", "INT", "REAL"]

    datum[2] = df[["ActorID", "FilmID"]].rename(
        columns={"ActorID": "start_id", "FilmID": "end_id"}
    )
    datum[2].insert(0, "serial", range(1, len(datum[2]) + 1))
    types[2] = ["SERIAL", "TEXT", "TEXT"]

    with pg.connect(con_string) as conn:
        with conn.cursor() as cur:
            for idx, (table, data, type) in enumerate(
                zip(src_tables.values(), datum, types)
            ):
                cur.execute(f"DROP TABLE IF EXISTS {table}")
                cols = ",".join(
                    [
                        col + " " + tp
                        for _, (col, tp) in enumerate(zip(data.columns, type))
                    ]
                )
                cur.execute(f"CREATE TABLE {table} ({cols})")
                query = f"COPY {table} FROM STDIN (FORMAT TEXT, FREEZE)"
                with cur.copy(query) as copy:
                    copy.write(
                        "\n".join(
                            [
                                "\t".join(map(str, row))
                                for row in data.itertuples(index=False)
                            ]
                        )
                    )
            cur.execute("COMMIT")


async def test_loadFromParquet(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
    init_parquet: bool = False,
) -> None:
    """
    Test for loadFromParquet()
    create parquet from actorfilms.csv
    after creating parquet, load it to a graph
    """
    src_parquet = "../data/actorfilms.parquet"

    if init_parquet:
        pd.read_csv("../data/actorfilms.csv").to_parquet(src_parquet)

    graph_name = "actorfilmsFromParquet"
    start_v_label = "Actor"
    edge_type = "ACTED_IN"
    end_v_label = "Film"
    start_time = time.time()
    await af.loadFromParquet(
        src_parquet=src_parquet,
        graph_name=graph_name,
        start_v_label=start_v_label,
        start_id="ActorID",
        start_props=["Actor"],
        edge_type=edge_type,
        end_v_label=end_v_label,
        end_id="FilmID",
        end_props=["Film", "Year", "Votes", "Rating"],
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=["Actor", "Film"],
        vertex_counts=[9623, 44456],
        edge_type=edge_type,
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def test_loadFromCosmosGremlin(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
    init_gremlin: bool = True,
) -> None:
    """
    NEED MORE TUNING

    Test for loadFromCosmosGremlin()
    create graph via Gremlin from actorfilms.csv
    after creating graph, load it to a graph

    export COSMOS_GREMLIN_ENDPOINT='wss://account_name.gremlin.cosmos.azure.com:443/'
    export COSMOS_GREMLIN_KEY='OwA3fVHzGzs8LsTN...........'
    """
    try:
        cosmos_gremlin_endpoint = os.environ["COSMOS_GREMLIN_ENDPOINT"]
        cosmos_gremlin_key = os.environ["COSMOS_GREMLIN_KEY"]
    except KeyError:
        print(
            "Please set the environment variables COSMOS_GREMLIN_ENDPOINT / COSMOS_GREMLIN_KEY"
        )
        return

    cosmos_db_name = "db1"
    cosmos_graph_name = "actorfilms"
    cosmos_username = f"/dbs/{cosmos_db_name}/colls/{cosmos_graph_name}"
    cosmos_pkey = "pk"

    if init_gremlin:
        await loadTestDataViaGremlin(
            cosmos_gremlin_endpoint=cosmos_gremlin_endpoint,
            cosmos_gremlin_key=cosmos_gremlin_key,
            cosmos_username=cosmos_username,
            cosmos_pkey=cosmos_pkey,
            src_csv="../data/actorfilms.csv",
        )

    start_time = time.time()
    graph_name = "actorfilmsfromgremlin"
    await af.loadFromCosmosGremlin(
        cosmos_gremlin_endpoint=cosmos_gremlin_endpoint,
        cosmos_gremlin_key=cosmos_gremlin_key,
        cosmos_username=cosmos_username,
        cosmos_pkey=cosmos_pkey,
        graph_name=graph_name,
        id_map={"Actor": "ActorID", "Film": "FilmID"},
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=["Actor", "Film"],
        vertex_counts=[9623, 44456],
        edge_type="ACTED_IN",
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def loadTestDataViaGremlin(
    cosmos_gremlin_endpoint: str = "",
    cosmos_gremlin_key: str = "",
    cosmos_username: str = "",
    cosmos_pkey: str = "",
    src_csv: str = "",
) -> None:
    """
    Load test data to Cosmos Gremlin
    file downloaded from https://www.kaggle.com/datasets/darinhawley/imdb-films-by-actor-for-10k-actors
    """
    from gremlin_python.driver import client, serializer

    # edge are automaticaly located in the same partirion as the 'from' node
    # AVERAGE_SIZE_OF_DOCUMENT includes the estimated size of the edge document
    LOGICAL_PARTITION_SIZE = 20 * 1024 * 1024 * 1024  # 20GB
    AVERAGE_SIZE_OF_DOCUMENT = 512  # 512bytes
    num_of_documents_per_partition = LOGICAL_PARTITION_SIZE // AVERAGE_SIZE_OF_DOCUMENT

    try:
        g_client = client.Client(
            url=cosmos_gremlin_endpoint,
            traversal_source="g",
            username=cosmos_username,
            password=cosmos_gremlin_key,
            message_serializer=serializer.GraphSONSerializersV2d0(),
        )
    except Exception as e:
        print(f"Failed to connect to Gremlin server: {e}")
        return

    # g_client.submitAsync("g.V().drop()")

    num_of_pk = 1

    df = pd.read_csv(src_csv)
    actors = df[["Actor", "ActorID"]].drop_duplicates()
    actors = actors.map(lambda x: x.replace("'", r"\'") if isinstance(x, str) else x)
    total_num_of_documents = 0
    for idx, (actor, actorid) in actors.iterrows():
        query = "g.addV('Actor').property('Actor', '{actor}').property('ActorID', '{actorid}').property('{pk}', '{num_of_pk}')".format(
            actorid=actorid, actor=actor, pk=cosmos_pkey, num_of_pk=num_of_pk
        )
        g_client.submitAsync(query).result()
        total_num_of_documents += 1
        if total_num_of_documents % num_of_documents_per_partition == 0:
            num_of_pk += 1

    films = df[["Film", "FilmID", "Year", "Votes", "Rating"]].drop_duplicates()
    films = films.map(lambda x: x.replace("'", r"\'") if isinstance(x, str) else x)
    for idx, (film, filmid, year, votes, rating) in films.iterrows():
        query = "g.addV('Film').property('Film', '{film}').property('FilmID', '{filmid}').property('Year', {year}).property('Votes', {votes}).property('Rating', {rating}).property('{pk}', '{num_of_pk}')".format(
            filmid=filmid,
            film=film,
            year=year,
            votes=votes,
            rating=rating,
            pk=cosmos_pkey,
            num_of_pk=num_of_pk,
        )
        g_client.submitAsync(query).result()
        total_num_of_documents += 1
        if total_num_of_documents % num_of_documents_per_partition == 0:
            num_of_pk += 1

    # can not avoid cross-partition query when the total size of documents exceeds the maximum size of logical partition, 20GB, because actor and film are in different partition
    for row in df.itertuples(index=False):
        query = "g.V().has('ActorID', '{actorid}').addE('ACTED_IN').to(g.V().has('FilmID', '{filmid}'))".format(
            actorid=row.ActorID, filmid=row.FilmID
        )
        g_client.submitAsync(query).result()
    g_client.close()


async def test_loadFromAvro(
    af: AgeFreighter,
    chunk_size: int = 96,
    direct_loading: bool = False,
    use_copy: bool = False,
    init_avro: bool = False,
) -> None:
    """
    NOT IMPLEMENTED YET

    Test for loadFromAvro()
    create avro from actorfilms.csv
    after creating avro, load it to a graph
    """
    src_avro = "../data/actorfilms.avro"

    if init_avro:
        await convertCSVtoAvro(src_csv="../data/actorfilms.csv", tgt_avro=src_avro)

    graph_name = "actorfilmsfrom_avro"
    start_time = time.time()
    start_v_label = "Actor"
    edge_type = "ACTED_IN"
    end_v_label = "Film"
    await af.loadFromAvro(
        src_avro=src_avro,
        graph_name=graph_name,
        start_v_label=start_v_label,
        start_id="ActorID",
        start_props=["Actor"],
        edge_type=edge_type,
        end_v_label=end_v_label,
        end_id="FilmID",
        end_props=["Film", "Year", "Votes", "Rating"],
        chunk_size=chunk_size,
        direct_loading=direct_loading,
        drop_graph=True,
        use_copy=use_copy,
    )
    execution_time = time.time() - start_time
    if not await is_graph_created(
        af,
        graph_name=graph_name,
        vertex_labels=[start_v_label, end_v_label],
        vertex_counts=[9623, 44456],
        edge_type=edge_type,
        edge_count=191873,
    ):
        print(f"{sys._getframe().f_code.co_name} failed")
    else:
        await show_test_result(
            sys._getframe().f_code.co_name,
            execution_time,
            chunk_size,
            direct_loading,
            use_copy,
        )


async def convertCSVtoAvro(src_csv: str = "", tgt_avro: str = "") -> None:
    """
    Convert CSV to Avro
    """
    import fastavro as fa

    records = pd.read_csv(src_csv).to_dict(orient="records")
    schema = {
        "type": "record",
        "name": "actorfilms",
        "fields": [
            {"name": "ActorID", "type": "string"},
            {"name": "Actor", "type": "string"},
            {"name": "FilmID", "type": "string"},
            {"name": "Film", "type": "string"},
            {"name": "Year", "type": "int"},
            {"name": "Votes", "type": "int"},
            {"name": "Rating", "type": "float"},
        ],
    }
    parsed_schema = fa.parse_schema(schema)
    with open(tgt_avro, "wb") as f:
        fa.writer(f, parsed_schema, records)


async def main() -> None:
    """
    Test for AgeFreighter

    export PG_CONNECTION_STRING="host=your_server.postgres.database.azure.com port=5432 dbname=postgres user=account password=your_password"

    Strongly reccomend to adjust chunk_size with your data and server before loading large amount of data
    Especially, the number of properties in the vertex affects the complecity of the query
    Due to asynchronous nature of the library, the duration for loading data is not linear to the number of rows

    Addition to the chunk_size, max_wal_size and checkpoint_timeout in the postgresql.conf should be considered
    """
    try:
        connection_string = os.environ["PG_CONNECTION_STRING"]
    except KeyError:
        print("Please set the environment variable PG_CONNECTION_STRING")
        return

    try:
        af = await AgeFreighter.connect(dsn=connection_string, max_connections=64)
        test_set = [
            [False, False],
            [True, False],
            [False, True],
        ]

        chunk_size = 128
        do = True
        if do:
            [
                await test_loadFromSingleCSV(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print("test_loadFromSingleCSV done\n")

        do = True
        if do:
            [
                await test_loadFromCSVs(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print("test_loadFromCSVs done\n")

        do = True
        if do:
            [
                await test_loadFromNetworkx(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print("test_loadFromNetworkx done\n")

        do = True
        if do:
            [
                await test_loadFromNeo4j(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                    init_neo4j=True,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print(
                "test_loadFromNeo4j done\n"
                "##### The duration for test_loadFromNeo4j depends on the performance of the neo4j server. #####\n"
            )

        do = True
        if do:
            [
                await test_loadFromPGSQL(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                    init_pgsql=True,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print(
                "test_loadFromPGSQL done\n"
                "##### The duration for test_loadFromPGSQL depends on the performance of the source pgsql server. #####\n"
            )

        do = True
        if do:
            [
                await test_loadFromParquet(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                    init_parquet=True,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print("test_loadFromParquet done\n")

        # NEED MORE TUNING
        do = True
        if do:
            [
                await test_loadFromCosmosGremlin(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                    init_gremlin=False,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print(
                "test_loadFromCosmosGremlin done\n"
                "##### The duration for test_loadFromCosmosGremlin depends on the performance of the source Cosmos DB. #####\n"
            )

        # NOT IMPLEMENTED YET
        do = False
        if do:
            [
                await test_loadFromAvro(
                    af,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    use_copy=use_copy,
                    init_avro=True,
                )
                for idx, (direct_loading, use_copy) in enumerate(test_set)
            ]
            print("test_loadFromAvro done\n")

    finally:
        await af.pool.close()


if __name__ == "__main__":
    asyncio.run(main())



import logging
from networkx import DiGraph
import pandas as pd
from psycopg import sql
from psycopg.rows import namedtuple_row
from psycopg_pool import AsyncConnectionPool
from typing_extensions import Callable
import nest_asyncio
from neo4j import AsyncGraphDatabase

nest_asyncio.apply()

log = logging.getLogger("agefreighter")


class Neo4jHandler:
    def __init__(self, uri, user, password):
        self.driver = AsyncGraphDatabase.driver(uri, auth=(user, password))

    def close(self):
        self.driver.close()

    def get_edges_with_nodes(self, chunk_size):
        with self.driver.session() as session:
            skip = 0
            while True:
                result = session.run(
                    """
                    MATCH (start)-[r]->(end)
                    RETURN start, r, end
                    SKIP $skip LIMIT $limit
                    """,
                    skip=skip,
                    limit=chunk_size,
                )
                records = result.data()
                if not records:
                    break
                yield records
                skip += chunk_size


class AgeFreighter:
    """
    AgeFreighter is a Python package that helps you to create a graph database using Azure Database for PostgreSQL.
    """

    def __init__(self):
        log.info("Creating AgeFreighter.")
        self.pool: AsyncConnectionPool = None
        self.dsn: str = ""
        self.graph_name: str = ""
        self.name = "AgeFreighter"
        self.version = "0.4.4"
        self.author = "Rio Fujita"

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.pool.close()

    def checkKeys(keys: list = [], elements: list = []):
        """
        Check if the keys of the CSV file match the elements.

        Args:
            keys (list): The keys of the CSV file.
            elements (list): The elements to check.
        """
        import numpy as np

        if not np.all(np.isin(elements, keys)):
            raise ValueError(
                f"CSV file must have {elements} columns, but {keys} columns were found."
            )

    def quotedGraphName(graph_name: str = "") -> str:
        """
        Quote the graph name.

        Args:
            graph_name (str): The name of the graph.
        """
        if graph_name.lower() != graph_name:
            return f'"{graph_name}"'
        return graph_name

    async def setUpGraph(self, graph_name: str = "", drop_graph: bool = False) -> None:
        """
        Set up the graph in the PostgreSQL database.

        Args:
            graph_name (str): The name of the graph to set up.
            drop_graph (bool): Whether to drop the existing graph if it exists.

        Returns:
            None
        """
        log.info("Setting up graph '%s'", graph_name)
        # for more precise graph name,
        # see https://github.com/apache/age/blob/master/src/include/utils/name_validation.h
        self.graph_name = sql.Identifier(graph_name).as_string().strip('"')
        log.info("self.graph_name is '%s'", self.graph_name)
        if drop_graph:
            async with self.pool.connection() as conn:
                async with conn.cursor(row_factory=namedtuple_row) as cur:
                    await cur.execute(
                        sql.SQL("CREATE EXTENSION IF NOT EXISTS age CASCADE")
                    )
                    await cur.execute(
                        sql.SQL(
                            f"SELECT count(*) FROM ag_graph WHERE name='{self.graph_name}'"
                        )
                    )
                    if (row := await cur.fetchone()) is not None:
                        if row.count == 1:
                            await cur.execute(
                                sql.SQL(f"SELECT drop_graph('{self.graph_name}', true)")
                            )
                        await cur.execute(
                            sql.SQL(f"SELECT create_graph('{self.graph_name}')")
                        )

    async def createLabelType(self, label_type: str = "", value: str = "") -> None:
        """
        Create a label type in the PostgreSQL database.

        Args:
            label_type (str): The type of the label to create. It can be either "vertex" or "edge".
            value (str): The value of the label to create.

        Returns:
            None
        """
        async with self.pool.connection() as conn:
            async with conn.cursor() as cur:
                if label_type == "vertex":
                    log.debug("Creating a vlabel '%s'", value)
                    await cur.execute(
                        sql.SQL("SELECT create_vlabel({schema}, {label});").format(
                            schema=self.graph_name,
                            label=sql.Literal(value),
                        )
                    )
                    log.debug("Creating index using GIN")
                    await cur.execute(
                        sql.SQL(
                            "CREATE INDEX ON {schema}.{label} USING GIN (properties);"
                        ).format(
                            schema=sql.Identifier(self.graph_name),
                            label=sql.Identifier(value),
                        )
                    )
                    log.debug("Creating index using BTREE")
                    await cur.execute(
                        sql.SQL(
                            "CREATE INDEX ON {schema}.{label} USING BTREE (id);"
                        ).format(
                            schema=sql.Identifier(self.graph_name),
                            label=sql.Identifier(value),
                        )
                    )
                elif label_type == "edge":
                    log.debug("Creating an elabel '%s'", value)
                    await cur.execute(
                        sql.SQL("SELECT create_elabel({schema}, {label});").format(
                            schema=self.graph_name,
                            label=sql.Literal(value),
                        )
                    )
                    log.debug("Creating index using BTREE")
                    await cur.execute(
                        sql.SQL("CREATE INDEX ON {schema}.{label} (start_id);").format(
                            schema=sql.Identifier(self.graph_name),
                            label=sql.Identifier(value),
                        )
                    )
                    log.debug("Creating index using BTREE")
                    await cur.execute(
                        sql.SQL("CREATE INDEX ON {schema}.{label} (end_id);").format(
                            schema=sql.Identifier(self.graph_name),
                            label=sql.Identifier(value),
                        )
                    )

    async def createVertices(
        self,
        vertices: pd.DataFrame = None,
        vertex_label: str = "",
        chunk_size: int = 3,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = False,
    ) -> None:
        """
        Create vertices in the PostgreSQL database.

        Args:
            vertices (pd.DataFrame): The vertices to create.
            vertex_label (str): The label of the vertices.
            chunk_size (int): The size of the chunks to create.
            direct_loading (bool): Whether to load the vertices directly.
            drop_graph (bool): Whether to drop the existing graph if it exists.
            use_copy (bool): Whether to use the COPY protocol to load the vertices.

        Returns:
            None
        """
        log.info("Creating vertices")
        log.debug("Number of vertices to be created: '%s'", len(vertices))
        if direct_loading:
            vertices = vertices.map(
                lambda x: x.replace("'", "''") if isinstance(x, str) else x
            )
            await self.createVerticesDirectly(self, vertices, vertex_label, chunk_size)
        else:
            vertices = vertices.map(
                lambda x: x.replace("'", r"\'") if isinstance(x, str) else x
            )
            if use_copy:
                await self.copyVertices(
                    self, vertices, vertex_label, chunk_size, drop_graph
                )
            else:
                await self.createVerticesCypher(
                    self, vertices, vertex_label, chunk_size
                )

    async def createEdges(
        self,
        edges: pd.DataFrame = None,
        edge_type: str = "",
        chunk_size: int = 3,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = False,
    ) -> None:
        """
        Create edges in the PostgreSQL database.

        Args:
            edges (pd.DataFrame): The edges to create.
            edge_type (str): The type of the edges.
            chunk_size (int): The size of the chunks to create.
            direct_loading (bool): Whether to load the edges directly.
            drop_graph (bool): Whether to drop the existing graph if it exists.
            use_copy (bool): Whether to use the COPY protocol to load the edges.

        Returns:
            None
        """
        log.info("Creating edges")
        log.debug("Number of edges to be created: '%s'", len(edges))
        # create edges
        if direct_loading:
            await self.createEdgesDirectly(self, edges, edge_type, chunk_size)
        else:
            if use_copy:
                await self.copyEdges(self, edges, edge_type, chunk_size)
            else:
                await self.createEdgesCypher(self, edges, edge_type, chunk_size)

    async def createVerticesCypher(
        self, vertices: pd.DataFrame = None, label: str = "", chunk_size: int = 0
    ) -> None:
        """
        Create vertices in the PostgreSQL database via Cypher.

        Args:
            vertices (pd.DataFrame): The vertices to create.
            label (str): The label of the vertices.
            chunk_size (int): The size of the chunks to create.

        Returns:
            None
        """
        log.info("Creating vertices via Cypher")
        chunk_multiplier = 2
        args = []
        for i in range(0, len(vertices), chunk_size * chunk_multiplier):
            chunk = vertices.iloc[i : i + chunk_size * chunk_multiplier]
            parts = chunk.apply(
                lambda row: "(v{0}:{1} {{{2}}})".format(
                    row.name,
                    label,
                    ",".join([f"{k}:'{v}'" for k, v in row.items()]),
                ),
                axis=1,
            ).tolist()
            args.append(
                f"SELECT * FROM cypher('{self.graph_name}', $$ CREATE {','.join(parts)} $$) AS (a agtype);"
            )
        await self.executeWithTasks(self, self.executeQuery, args)

    async def createEdgesCypher(
        self, edges: pd.DataFrame = None, type: str = "", chunk_size: int = 0
    ) -> None:
        """
        Create edges in the PostgreSQL database via Cypher.

        Args:
            edges (pd.DataFrame): The edges to create.
            type (str): The type of the edges.
            chunk_size (int): The size of the chunks to create.

        Returns:
            None
        """
        log.info("Creating edges via Cypher")
        chunk_multiplier = 2
        args = []
        for i in range(0, len(edges), chunk_size * chunk_multiplier):
            chunk = edges.iloc[i : i + chunk_size * chunk_multiplier]
            parts = chunk.apply(
                lambda row: (
                    f"MATCH (n:{row['start_v_label']} {{id: '{row['start_id']}'}}), "
                    f"(m:{row['end_v_label']} {{id: '{row['end_id']}'}}) "
                    f"CREATE (n)-[:{type}]->(m)"
                ),
                axis=1,
            ).tolist()

            query = "".join(
                [
                    f"SELECT * FROM cypher('{self.graph_name}', $$ {part} $$) AS (a agtype);"
                    for part in parts
                ]
            )
            log.debug("Query to be executed: '%s'", query)
            args.append(query)
        await self.executeWithTasks(self, self.executeQuery, args)

    async def createVerticesDirectly(
        self, vertices: pd.DataFrame = None, label: str = "", chunk_size: int = 0
    ) -> None:
        """
        Create vertices in the PostgreSQL database directly.

        Args:
            vertices (pd.DataFrame): The vertices to create.
            label (str): The label of the vertices.
            chunk_size (int): The size of the chunks to create.

        Returns:
            None
        """
        log.info("Creating vertices with SQL query")
        chunk_multiplier = 2
        args = []
        graph_name = self.quotedGraphName(self.graph_name)
        for i in range(0, len(vertices), chunk_size * chunk_multiplier):
            chunk = vertices.iloc[i : i + chunk_size * chunk_multiplier]
            values = chunk.apply(
                lambda row: "('{"
                + ",".join([f'"{k}":"{v}"' for k, v in row.items()])
                + "}')",
                axis=1,
            ).tolist()
            args.append(
                f"INSERT INTO {graph_name}.\"{label}\" (properties) VALUES {','.join(values)};"
            )
        await self.executeWithTasks(self, self.executeQuery, args)

    async def createEdgesDirectly(
        self, edges: pd.DataFrame = None, type: str = "", chunk_size: int = 0
    ) -> None:
        """
        Create edges in the PostgreSQL database directly.

        Args:
            edges (pd.DataFrame): The edges to create.
            type (str): The type of the edges.
            chunk_size (int): The size of the chunks to create.

        Returns:
            None
        """
        log.info("Creating edges with SQL query")
        chunk_multiplier = 2
        # create id_maps to convert entry_id to id(graphid)
        id_maps = await self.getIdMaps(self, edges=edges)
        log.debug("ID_maps: '%s'", id_maps)

        # create queries for edges
        args = []
        graph_name = self.quotedGraphName(self.graph_name)
        for i in range(0, len(edges), chunk_size * chunk_multiplier):
            chunk = edges.iloc[i : i + chunk_size * chunk_multiplier]
            values = chunk.apply(
                lambda row: f"('{id_maps[str(row['start_v_label'])][str(row['start_id'])]}'::graphid, '{id_maps[str(row['end_v_label'])][str(row['end_id'])]}'::graphid)",
                axis=1,
            ).tolist()
            args.append(
                f"INSERT INTO {graph_name}.\"{type}\" (start_id, end_id) VALUES {','.join(values)};"
            )
            log.debug("Query to be excecuted: '%s'", args)
        await self.executeWithTasks(self, self.executeQuery, args)

    async def executeWithTasks(self, target: Callable = None, args: list = []) -> None:
        """
        Execute queries with tasks.

        Args:
            target (Callable): The target function to execute.
            args (list): The arguments to pass to the target function.
        """
        import asyncio

        tasks = []
        for arg in args:
            task = asyncio.create_task(target(self.pool, arg))
            tasks.append(task)
        await asyncio.gather(*tasks)

    async def executeQuery(pool: AsyncConnectionPool = None, query: str = "") -> None:
        """
        Execute a query with an async connection pool.

        Args:
            pool (AsyncConnectionPool): The async connection pool to use.
            query (str): The query to execute.

        Returns:
            None
        """
        import asyncio

        while True:
            try:
                async with pool.connection() as conn:
                    async with conn.cursor() as cur:
                        await cur.execute(query)
                        break
            except Exception as e:
                print(e)
                await asyncio.sleep(1)
                pass

    async def copyVertices(
        self,
        vertices: pd.DataFrame = None,
        label: str = "",
        chunk_size: int = 0,
        drop_graph: bool = False,
    ) -> None:
        """
        Create vertices in the PostgreSQL database via the COPY protocol.

        Args:
            vertices (pd.DataFrame): The vertices to create.
            label (str): The label of the vertices.
            chunk_size (int): The size of the chunks to create.
            drop_graph (bool): Whether to drop the existing graph if it exists.

        Returns:
            None
        """
        log.info("Copying vertices via COPY protocol")
        chunk_multiplier = 1000
        first_id = await self.getFirstId(self, label)
        graph_name = self.quotedGraphName(self.graph_name)
        async with self.pool.connection() as conn:
            async with conn.cursor() as cur:
                query = f'COPY {graph_name}."{label}" FROM STDIN (FORMAT TEXT)'
                log.debug("Query to be excecuted: '%s'", query)
                async with cur.copy(query) as copy:
                    args_list = []
                    for i in range(0, len(vertices), chunk_size * chunk_multiplier):
                        chunk = vertices.iloc[i : i + chunk_size * chunk_multiplier]
                        chunk_args = chunk.apply(
                            lambda row: "{0}\t{{{1}}}\n".format(
                                first_id + row.name,
                                ", ".join([f'"{k}": "{v}"' for k, v in row.items()]),
                            ),
                            axis=1,
                        ).tolist()
                        args_list.extend(chunk_args)
                        args = "".join(args_list)
                        await copy.write(args)
                        log.debug("Args to be passed to COPY: '%s'", args)
                await cur.execute(sql.SQL("COMMIT"))

    async def copyEdges(
        self,
        edges: pd.DataFrame = None,
        type: str = "",
        chunk_size: int = 0,
    ) -> None:
        """
        Create edges in the PostgreSQL database via the COPY protocol.

        Args:
            edges (pd.DataFrame): The edges to create.
            type (str): The type of the edges.
            chunk_size (int): The size of the chunks to create.
            drop_graph (bool): Whether to drop the existing graph if it exists.

        Returns:
            None
        """
        log.info("Copying vertices via COPY protocol")
        chunk_multiplier = 1000

        # create id_maps to convert entry_id to id(graphid)
        id_maps = await self.getIdMaps(self, edges=edges)
        log.debug("IDmaps: '%s'", id_maps)

        # create queries for edges
        first_id = await self.getFirstId(self, label_type=type)
        graph_name = self.quotedGraphName(self.graph_name)
        async with self.pool.connection() as conn:
            async with conn.cursor() as cur:
                query = f'COPY {graph_name}."{type}" (id,start_id,end_id) FROM STDIN (FORMAT TEXT)'
                log.debug("Query to be excecuted: '%s'", query)
                async with cur.copy(query) as copy:
                    for i in range(0, len(edges), chunk_size * chunk_multiplier):
                        args_list = []
                        chunk = edges.iloc[i : i + chunk_size * chunk_multiplier]
                        chunk_args = chunk.apply(
                            lambda row: "{0}\t{1}\t{2}\n".format(
                                first_id + row.name,
                                id_maps[str(row["start_v_label"])][
                                    str(row["start_id"])
                                ],
                                id_maps[str(row["end_v_label"])][str(row["end_id"])],
                            ),
                            axis=1,
                        ).tolist()
                        args_list.extend(chunk_args)
                        args = "".join(args_list)
                        await copy.write(args)
                        log.debug("Args to be passed to COPY: '%s'", args)

    async def getIdMaps(self, edges: pd.DataFrame = None) -> dict:
        """
        Get the idmaps between entry_id and id(graphid).

        Args:
            edges (pd.DataFrame): The edges to create.

        Returns:
            dict: The ID maps.
        """
        # create id_maps to convert entry_id to id(graphid)
        id_maps = {}
        graph_name = self.quotedGraphName(self.graph_name)
        async with self.pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                for e_label in [
                    edges.iloc[0]["start_v_label"],
                    edges.iloc[0]["end_v_label"],
                ]:
                    await cur.execute(
                        sql.SQL(
                            f'SELECT id, properties->\'"id"\' AS entry_id FROM {graph_name}."{e_label}"'
                        )
                    )
                    rows = await cur.fetchall()
                    id_maps[e_label] = {
                        row.entry_id.replace('"', ""): row.id for row in rows
                    }

        return id_maps

    async def getFirstId(self, label_type: str = "") -> int:
        """
        Get the first id for a vertex or edge.

        Args:
            label_type (str): The type of the label to get the first id for.

        Returns:
            int: The first id.
        """
        import numpy as np

        async with self.pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                await cur.execute(
                    sql.SQL(f"SELECT id FROM ag_label WHERE name='{label_type}'")
                )
                row = await cur.fetchone()

                ENTRY_ID_BITS = 32 + 16
                ENTRY_ID_MASK = np.uint64(0x0000FFFFFFFFFFFF)
                first_id = ((np.uint64(row.id)) << ENTRY_ID_BITS) | (
                    (np.uint64(1)) & ENTRY_ID_MASK
                )

                return first_id

    async def renameColumns(df: pd.DataFrame = None) -> pd.DataFrame:
        """
        Rename the columns of a DataFrame.

        Args:
            df (pd.DataFrame): The DataFrame to rename the columns of.

        Returns:
            pd.DataFrame: The DataFrame with the renamed columns
        """
        cols_to_rename = {}
        for col in df.columns.tolist():
            if col.startswith("a().prop."):
                cols_to_rename[col] = col.replace("a().prop.", "")
            elif col.startswith("a()."):
                cols_to_rename[col] = col.replace("a().", "")
            if col.startswith("a->.prop."):
                cols_to_rename[col] = col.replace("a->.prop.", "")
            elif col.startswith("a->."):
                cols_to_rename[col] = col.replace("a->.", "")

        return df.rename(columns=cols_to_rename)

    async def getVertexLabels(
        nodes: pd.DataFrame = None,
        vertex_labels: list = [],
        first_source_id: str = "",
        first_target_id: str = "",
    ) -> list:
        """
        Get the start and end vertex labels.

        Args:
            nodes (pd.DataFrame): The nodes to get the start and end vertex labels from.
            vertex_labels (list): The vertex labels.
            first_source_id (str): The first source id.
            first_target_id (str): The first target id.

        Returns:
            list: The start and end vertex labels.
        """
        try:
            start_v_label = [
                item["label"] for item in nodes if item["id"] == first_source_id
            ][0]
            end_v_label = (
                vertex_labels[1]
                if vertex_labels[0] == start_v_label
                else vertex_labels[0]
            )
        except IndexError:
            end_v_label = [
                item["label"] for item in nodes if item["id"] == first_target_id
            ][0]
            start_v_label = (
                vertex_labels[1]
                if vertex_labels[0] == end_v_label
                else vertex_labels[0]
            )

        return start_v_label, end_v_label

    def getChunks(df: pd.DataFrame = None, chunk_size: int = 0) -> pd.DataFrame:
        """
        Get the DataFrame in chunks.

        Args:
            df (DataFrame): The DataFrame to get the edges from.
            chunk_size (int): The size of the chunks to get the edges in.

        Returns:
            DataFrame: chunk of the DataFrame
        """
        for i in range(0, len(df), chunk_size):
            yield df.iloc[i : i + chunk_size].copy()

    def get_node_attributes(graph, node_ids, attributes):
        """
        Get node attributes.

        Args:
            graph: The graph.
            node_ids: The node IDs.
            attributes: The attributes.

        Returns:
            dict: The node attributes.
        """
        return {
            attr: [graph.nodes[node_id].get(attr, None) for node_id in node_ids]
            for attr in attributes
        }

    @classmethod
    async def connect(
        cls, dsn: str = "", max_connections: int = 64, log_level=None, **kwargs
    ) -> "AgeFreighter":
        """
        Open a connection pool.

        Args:
            dsn (str): The data source name.
            max_connections (int): The maximum number of connections.
            log_level: The log level.
            **kwargs: Additional keyword arguments.

        Returns:
            AgeFreighter: The AgeFreighter object.
        """
        log.info("Opening connection pool")
        # to make large number of connections
        import resource

        current_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
        resource.setrlimit(resource.RLIMIT_NOFILE, (8192, current_limit[1]))

        cls.dsn = dsn + " options='-c search_path=ag_catalog,\"$user\",public'"
        cls.pool = AsyncConnectionPool(
            cls.dsn,
            max_size=max_connections,
            min_size=64,
            open=False,
            timeout=600,
            **kwargs,
        )
        log.debug("Connection pool: '%s'", cls.pool)
        await cls.pool.open()

        return cls

    async def createGraphFromDataFrame(
        cls,
        graph_name: str = "",
        src: pd.DataFrame = None,
        existing_node_ids: list = [],
        first_chunk: bool = True,
        start_v_label: str = "",
        start_id: str = "",
        start_props: list = [],
        edge_type: str = "",
        end_v_label: str = "",
        end_id: str = "",
        end_props: list = [],
        chunk_size: int = 3,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = False,
    ) -> None:
        """
        Create a graph from DataFrame

        Args:
            graph_name (str): The name of the graph to load the data into.
            src (pd.DataFrame): The DataFrame to load the data from.
            existing_node_ids (list): The existing node IDs.
            first_chunk (bool): Whether it is the first chunk.
            start_v_label (str): The label of the start vertex.
            start_id (str): The ID of the start vertex.
            start_props (list): The properties of the start vertex.
            edge_type (str): The type of the edge.
            end_v_label (str): The label of the end vertex.
            end_id (str): The ID of the end vertex.
            end_props (list): The properties of the end vertex.

        Returns:
            None
        """
        log.info("Creating a graph from DataFrame")
        if first_chunk:
            cls.checkKeys(
                src.keys(),
                [start_id] + start_props + [end_id] + end_props,
            )

            await cls.setUpGraph(cls, graph_name, drop_graph)

            await cls.createLabelType(cls, label_type="vertex", value=start_v_label)
            await cls.createLabelType(cls, label_type="vertex", value=end_v_label)
            await cls.createLabelType(cls, label_type="edge", value=edge_type)

        for vertex_label, id, props in zip(
            [start_v_label, end_v_label],
            [start_id, end_id],
            [start_props, end_props],
        ):
            vertices = (
                src.loc[:, [id, *props]].drop_duplicates().rename(columns={id: "id"})
            )
            if not first_chunk:
                vertices = vertices[~vertices["id"].isin(existing_node_ids)]
            existing_node_ids.extend(vertices["id"].tolist())
            await cls.createVertices(
                cls,
                vertices,
                vertex_label,
                chunk_size,
                direct_loading,
                drop_graph,
                use_copy,
            )

        edges = (
            src.loc[:, [start_id, end_id]]
            .drop_duplicates()
            .rename(columns={start_id: "start_id", end_id: "end_id"})
        )
        edges["start_v_label"] = start_v_label
        edges["end_v_label"] = end_v_label
        await cls.createEdges(
            cls, edges, edge_type, chunk_size, direct_loading, drop_graph, use_copy
        )

    @classmethod
    async def loadFromSingleCSV(
        cls,
        csv: str = "",
        graph_name: str = "",
        start_v_label: str = "",
        start_id: str = "",
        start_props: list = [],
        edge_type: str = "",
        end_v_label: str = "",
        end_id: str = "",
        end_props: list = [],
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a single CSV file.

        Args:
            csv (str): The path to the CSV file.
            graph_name (str): The name of the graph to load the data into.
            start_v_label (str): The label of the start vertex.
            start_id (str): The ID of the start vertex.
            start_props (list): The properties of the start vertex.
            edge_type (str): The type of the edge.
            end_v_label (str): The label of the end vertex.
            end_id (str): The ID of the end vertex.
            end_props (list): The properties of the end vertex.
            chunk_size (int): The size of the chunks to create.
            direct_loading (bool): Whether to load the data directly.
            drop_graph (bool): Whether to drop the existing graph if it exists.
            use_copy (bool): Whether to use the COPY protocol to load the data.

        Returns:
            None
        """
        log.info("Loading data from a single CSV file")
        chunk_multiplier = 10000
        existing_node_ids = []
        first_chunk = True
        reader = pd.read_csv(csv, chunksize=chunk_size * chunk_multiplier)
        for df in reader:
            await cls.createGraphFromDataFrame(
                cls,
                graph_name=graph_name,
                src=df,
                existing_node_ids=existing_node_ids,
                first_chunk=first_chunk,
                start_v_label=start_v_label,
                start_id=start_id,
                start_props=start_props,
                edge_type=edge_type,
                end_v_label=end_v_label,
                end_id=end_id,
                end_props=end_props,
                chunk_size=chunk_size,
                direct_loading=direct_loading,
                drop_graph=drop_graph,
                use_copy=use_copy,
            )
            first_chunk = False

    @classmethod
    async def loadFromCSVs(
        cls,
        vertex_csvs: list = [],
        vertex_labels: list = [],
        edge_csvs: list = [],
        edge_types: list = [],
        graph_name: str = "",
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from multiple CSV files.

        Args:
            graph_name (str): The name of the graph to load the data into.
            vertex_csvs (list): The paths to the vertex CSV files.
            vertex_labels (list): The labels of the vertices.
            edge_csvs (list): The paths to the edge CSV files.
            edge_types (list): The types of the edges.

        Returns:
            None
        """
        log.info("Loading data from multiple CSV files")
        chunk_multiplier = 10000
        await cls.setUpGraph(cls, graph_name, drop_graph)
        for vertex_csv, vertex_label in zip(vertex_csvs, vertex_labels):
            first_chunk = True
            reader = pd.read_csv(vertex_csv, chunksize=chunk_size * chunk_multiplier)
            for vertices in reader:
                if first_chunk:
                    cls.checkKeys(vertices.keys(), ["id"])

                    await cls.createLabelType(
                        cls, label_type="vertex", value=vertex_label
                    )
                    first_chunk = False

                await cls.createVertices(
                    cls,
                    vertices,
                    vertex_label,
                    chunk_size,
                    direct_loading,
                    drop_graph,
                    use_copy,
                )

        for edge_csv, edge_type in zip(edge_csvs, edge_types):
            first_chunk = True
            reader = pd.read_csv(edge_csv, chunksize=chunk_size * chunk_multiplier)
            for edges in reader:
                if first_chunk:
                    cls.checkKeys(
                        edges.keys(),
                        [
                            "start_id",
                            "start_vertex_type",
                            "end_id",
                            "end_vertex_type",
                        ],
                    )

                    await cls.createLabelType(cls, label_type="edge", value=edge_type)
                    first_chunk = False
                edges.rename(
                    columns={
                        "start_vertex_type": "start_v_label",
                        "end_vertex_type": "end_v_label",
                    },
                    inplace=True,
                )
                await cls.createEdges(
                    cls,
                    edges,
                    edge_type,
                    chunk_size,
                    direct_loading,
                    drop_graph,
                    use_copy,
                )

    @classmethod
    async def loadFromNetworkx(
        cls,
        networkx_graph: DiGraph = None,
        graph_name: str = "",
        id_map: dict = {},
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a NetworkX graph.

        Args:
            networkx_graph (DiGraph): The NetworkX graph.
            graph_name (str): The name of the graph to load the data into.
            id_map (dict): The ID map.

        Returns:
            None
        """
        log.info("Loading data from a NetworkX graph")
        import networkx as nx

        chunk_multiplier = 1000

        first_chunk = True
        existing_node_ids = []
        edges = nx.to_pandas_edgelist(networkx_graph)
        start_v_label = networkx_graph.nodes[edges.iloc[0]["source"]]["label"]
        start_id = id_map[start_v_label]
        end_v_label = networkx_graph.nodes[edges.iloc[0]["target"]]["label"]
        end_id = id_map[end_v_label]
        edge_type = edges.iloc[0]["label"]
        start_props = [
            prop
            for prop in networkx_graph.nodes[edges.iloc[0]["source"]]
            if prop not in ["label", "name"]
        ]
        end_props = [
            prop
            for prop in networkx_graph.nodes[edges.iloc[0]["target"]]
            if prop not in ["label", "name"]
        ]
        for chunk in cls.getChunks(edges, chunk_size * chunk_multiplier):
            source_ids = chunk["source"].tolist()
            target_ids = chunk["target"].tolist()

            start_attributes = cls.get_node_attributes(
                networkx_graph, source_ids, ["name"] + start_props
            )
            end_attributes = cls.get_node_attributes(
                networkx_graph, target_ids, ["name"] + end_props
            )

            chunk[start_v_label] = start_attributes["name"]
            for start_prop in start_props:
                chunk[start_prop] = start_attributes[start_prop]

            chunk[end_v_label] = end_attributes["name"]
            for end_prop in end_props:
                chunk[end_prop] = end_attributes[end_prop]

            chunk.rename(columns={"source": start_id, "target": end_id}, inplace=True)
            chunk.drop(columns=["label"], inplace=True)
            await cls.createGraphFromDataFrame(
                cls,
                graph_name=graph_name,
                src=chunk,
                existing_node_ids=existing_node_ids,
                first_chunk=first_chunk,
                start_v_label=start_v_label,
                start_id=start_id,
                start_props=start_props,
                edge_type=edge_type,
                end_v_label=end_v_label,
                end_id=end_id,
                end_props=end_props,
                chunk_size=chunk_size,
                direct_loading=direct_loading,
                drop_graph=drop_graph,
                use_copy=use_copy,
            )
            first_chunk = False

    @classmethod
    async def loadFromNeo4jNew(
        cls,
        uri: str = "bolt://localhost:7687",
        user: str = "neo4j",
        password: str = "neo4jpass",
        neo4j_database: str = "neo4j",
        graph_name: str = "",
        id_map: dict = {},
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a Neo4j graph.

        Args:
            uri (str): The URI of the Neo4j database.
            user (str): The user of the Neo4j database.
            password (str): The password of the Neo4j database.
            neo4j_database (str): The name of the Neo4j database.
            graph_name (str): The name of the graph to load the data into.
            id_map (dict): The ID map.

        Returns:
            None
        """
        log.info("Loading data from a Neo4j graph")

        chunk_multiplier = 100
        neo4j_handler = Neo4jHandler(uri, user, password)
        try:
            for chunk in neo4j_handler.get_edges_with_nodes(
                chunk_size * chunk_multiplier
            ):
                # チャンクごとに処理を行う
                for record in chunk:
                    print(record)
                    start_node = record["start"]
                    edge = record["r"]
                    end_node = record["end"]
                    # ここでstart_node, edge, end_nodeのプロパティを処理する
        finally:
            neo4j_handler.close()

    @classmethod
    async def loadFromNeo4j(
        cls,
        uri: str = "bolt://localhost:7687",
        user: str = "neo4j",
        password: str = "neo4jpass",
        neo4j_database: str = "neo4j",
        graph_name: str = "",
        id_map: dict = {},
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a Neo4j graph.

        Args:
            uri (str): The URI of the Neo4j database.
            user (str): The user of the Neo4j database.
            password (str): The password of the Neo4j database.
            neo4j_database (str): The name of the Neo4j database.
            graph_name (str): The name of the graph to load the data into.
            id_map (dict): The ID map.

        Returns:
            None
        """
        log.info("Loading data from a Neo4j graph")
        import neo4j

        await cls.setUpGraph(cls, graph_name, drop_graph)

        chunk_multiplier = 100
        async with neo4j.AsyncGraphDatabase.driver(
            uri, auth=(user, password)
        ) as driver:
            records, _, _ = await driver.execute_query(
                "MATCH (n) RETURN distinct labels(n)",
                db=neo4j_database,
            )
            vertex_labels = []
            for record in records:
                vertex_label = record["labels(n)"][0]
                vertex_labels.append(vertex_label)
                await cls.createLabelType(cls, label_type="vertex", value=vertex_label)
                result = await driver.execute_query(
                    f"MATCH (a:{vertex_label}) RETURN count(a) AS count",
                    db=neo4j_database,
                    result_transformer_=neo4j.AsyncResult.single,
                )
                cnt = result["count"]
                for i in range(0, cnt, chunk_size * chunk_multiplier):
                    vertices = await driver.execute_query(
                        f"MATCH (a:{vertex_label}) RETURN a SKIP $skip LIMIT $limit",
                        skip=i,
                        limit=chunk_size * chunk_multiplier,
                        db=neo4j_database,
                        result_transformer_=lambda res: res.to_df(expand=True),
                    )
                    vertices.drop(columns=["a().labels"], inplace=True)
                    vertices = await cls.renameColumns(vertices)
                    vertices.rename(columns={id_map[vertex_label]: "id"}, inplace=True)
                    await cls.createVertices(
                        cls,
                        vertices,
                        vertex_label,
                        chunk_size,
                        direct_loading,
                        drop_graph,
                        use_copy,
                    )

            records, _, _ = await driver.execute_query(
                "MATCH ()-[n]->() RETURN distinct type(n)",
                db=neo4j_database,
            )
            for record in records:
                edge_type = record["type(n)"]
                await cls.createLabelType(cls, label_type="edge", value=edge_type)
                # need to know the number of edges
                result = await driver.execute_query(
                    f"MATCH ()-[a:{edge_type}]->() RETURN count(a) AS count",
                    result_transformer_=neo4j.AsyncResult.single,
                )
                cnt = result["count"]
                result = await driver.execute_query(
                    f"MATCH (m) - [a:{edge_type}] -> (n) RETURN m, n LIMIT 1",
                    result_transformer_=lambda res: res.to_df(expand=True),
                )
                # fix me if the nodes have multiple labels
                start_v_label = "".join(map(str, result["m().labels"][0]))
                end_v_label = "".join(map(str, result["n().labels"][0]))
                for i in range(0, cnt, chunk_size * chunk_multiplier):
                    edges = await driver.execute_query(
                        f"MATCH () - [a:{edge_type}] -> () RETURN a SKIP $skip LIMIT $limit",
                        skip=i,
                        limit=chunk_size * chunk_multiplier,
                        db=neo4j_database,
                        result_transformer_=lambda res: res.to_df(expand=True),
                    )
                    edges.drop(columns=["a->.type"], inplace=True)
                    edges = await cls.renameColumns(edges)
                    edges.rename(
                        columns={"from": "start_id", "to": "end_id"}, inplace=True
                    )
                    edges.insert(0, "start_v_label", start_v_label)
                    edges.insert(0, "end_v_label", end_v_label)
                    await cls.createEdges(
                        cls,
                        edges,
                        edge_type,
                        chunk_size,
                        direct_loading,
                        drop_graph,
                        use_copy,
                    )

    @classmethod
    async def loadFromPGSQL(
        cls,
        src_con_string: str = "",
        src_tables: list = [],
        graph_name: str = "",
        id_map: dict = {},
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a PostgreSQL database.

        Args:
            src_con_string (str): The connection string of the source PostgreSQL database.
            src_tables (list): The source tables.
            graph_name (str): The name of the graph to load the data into.
            id_map (dict): The ID map.

        Returns:
            None
        """
        log.info("Loading data from a PostgreSQL database")
        import psycopg as pg

        chunk_multiplier = 1000

        try:
            with pg.connect(src_con_string) as conn:
                with conn.cursor(row_factory=namedtuple_row) as cur:
                    await cls.setUpGraph(cls, graph_name, drop_graph)
                    for src_table in src_tables.values():
                        if id_map.get(src_table) is not None:  # nodes
                            id_col_name = id_map[src_table]
                            await cls.createLabelType(
                                cls, label_type="vertex", value=src_table
                            )
                            cur.execute(sql.SQL(f"SELECT COUNT(*) FROM {src_table}"))
                            cnt = cur.fetchone()[0]
                            for i in range(0, cnt, chunk_size * chunk_multiplier):
                                cur.execute(
                                    sql.SQL(
                                        f"SELECT * FROM {src_table} LIMIT {chunk_size * chunk_multiplier} OFFSET {i}"
                                    )
                                )
                                rows = cur.fetchall()
                                vertices = pd.DataFrame(rows)
                                vertices.rename(
                                    columns={id_col_name: "id"}, inplace=True
                                )
                                await cls.createVertices(
                                    cls,
                                    vertices,
                                    src_table,
                                    chunk_size,
                                    direct_loading,
                                    drop_graph,
                                    use_copy,
                                )
                        else:  # edges
                            await cls.createLabelType(
                                cls, label_type="edge", value=src_table
                            )
                            cur.execute(sql.SQL(f"SELECT COUNT(*) FROM {src_table}"))
                            cnt = cur.fetchone()[0]
                            for i in range(0, cnt, chunk_size * chunk_multiplier):
                                cur.execute(
                                    sql.SQL(
                                        f"SELECT * FROM {src_table} LIMIT {chunk_size * chunk_multiplier} OFFSET {i}"
                                    )
                                )
                                rows = cur.fetchall()
                                edges = pd.DataFrame(rows)
                                edges.insert(0, "start_v_label", list(id_map.keys())[0])
                                edges.insert(0, "end_v_label", list(id_map.keys())[1])
                                await cls.createEdges(
                                    cls,
                                    edges,
                                    src_table,
                                    chunk_size,
                                    direct_loading,
                                    drop_graph,
                                    use_copy,
                                )
        except Exception as e:
            raise e

    @classmethod
    async def loadFromParquet(
        cls,
        src_parquet: str = "",
        graph_name: str = "",
        start_v_label: str = "",
        start_id: str = "",
        start_props: list = [],
        edge_type: str = "",
        end_v_label: str = "",
        end_id: str = "",
        end_props: list = [],
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a Parquet file.

        Args:
            src_parquet (str): The path to the Parquet file.

        Returns:
            None
        """
        log.info("Loading data from a Parquet file")
        from pyarrow.parquet import ParquetFile

        chunk_multiplier = 10000

        pf = ParquetFile(src_parquet)
        first_chunk = True
        existing_node_ids = []

        for batch in pf.iter_batches(chunk_size * chunk_multiplier):
            df = batch.to_pandas()
            await cls.createGraphFromDataFrame(
                cls,
                graph_name=graph_name,
                src=df,
                existing_node_ids=existing_node_ids,
                first_chunk=first_chunk,
                start_v_label=start_v_label,
                start_id=start_id,
                start_props=start_props,
                edge_type=edge_type,
                end_v_label=end_v_label,
                end_id=end_id,
                end_props=end_props,
                chunk_size=chunk_size,
                direct_loading=direct_loading,
                drop_graph=drop_graph,
                use_copy=use_copy,
            )
            first_chunk = False

    @classmethod
    async def loadFromAvro(
        cls,
        src_avro: str = "",
        graph_name: str = "",
        start_v_label: str = "",
        start_id: str = "",
        start_props: list = [],
        edge_type: str = "",
        end_v_label: str = "",
        end_id: str = "",
        end_props: list = [],
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from an Avro file.

        Args:
            src_avro (str): The path to the Avro file.

        Returns:
            None
        """
        log.info("Loading data from an Avro file")
        import fastavro as fa

        chunk_multiplier = 10000

        with open(src_avro, "rb") as f:
            reader = fa.reader(f)
            first_chunk = True
            existing_node_ids = []

            for records in reader:
                df = pd.DataFrame.from_records(records, index=[0])
                await cls.createGraphFromDataFrame(
                    cls,
                    graph_name=graph_name,
                    src=df,
                    existing_node_ids=existing_node_ids,
                    first_chunk=first_chunk,
                    start_v_label=start_v_label,
                    start_id=start_id,
                    start_props=start_props,
                    edge_type=edge_type,
                    end_v_label=end_v_label,
                    end_id=end_id,
                    end_props=end_props,
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    drop_graph=drop_graph,
                    use_copy=use_copy,
                )
                first_chunk = False

    @classmethod
    async def loadFromCosmosGremlin(
        cls,
        cosmos_gremlin_endpoint: str = "",
        cosmos_gremlin_key: str = "",
        cosmos_username: str = "",
        cosmos_pkey: str = "",
        graph_name: str = "",
        id_map: dict = {},
        chunk_size: int = 128,
        direct_loading: bool = False,
        drop_graph: bool = False,
        use_copy: bool = True,
    ) -> None:
        """
        Load data from a Gremlin graph.

        Args:
            cosmos_gremlin_endpoint (str): The Cosmos Gremlin endpoint.
            cosmos_gremlin_key (str): The Cosmos Gremlin key.
            cosmos_username (str): The Cosmos username.
            graph_name (str): The name of the graph to load the data into.
            id_map (dict): The ID map.

        Returns:
            None
        """
        log.info("Loading data from a Gremlin graph")
        from gremlin_python.driver import client, serializer

        chunk_multiplier = 100

        try:
            g = client.Client(
                url=cosmos_gremlin_endpoint,
                traversal_source="g",
                username=cosmos_username,
                password=cosmos_gremlin_key,
                message_serializer=serializer.GraphSONSerializersV2d0(),
            )
        except Exception as e:
            print(f"Failed to connect to Gremlin server: {e}")
            return

        query = "g.E().label().dedup()"
        log.debug("Query to be executed: '%s'", query)
        edge_types = [
            edge_type for edge_type in g.submit_async(query).result().all().result()
        ]

        for edge_type in edge_types:
            query = f"g.E().hasLabel('{edge_type}').count()"
            log.debug("Query to be executed: '%s'", query)
            cnt = g.submit_async(query).result().all().result()[0]
            query = f"g.E().hasLabel('{edge_type}').limit(1)"
            log.debug("Query to be executed: '%s'", query)
            vertex_labels = g.submit_async(query).result().all().result()
            in_v_label = vertex_labels[0]["inVLabel"]
            out_v_label = vertex_labels[0]["outVLabel"]
            existing_node_ids = []
            first_chunk = True
            for i in range(0, cnt, chunk_size * chunk_multiplier):
                query = f"g.E().hasLabel('{edge_type}').range({i}, {i + chunk_size * chunk_multiplier}).bothV()"
                log.debug("Query to be executed: '%s'", query)
                records = g.submit_async(query).result().all().result()
                dicts = []
                for j in range(0, len(records), 2):
                    dc = {}
                    for k, v in records[j]["properties"].items():
                        if k != "pk":
                            dc[k] = v[0]["value"]
                    for k, v in records[j + 1]["properties"].items():
                        if k != "pk":
                            dc[k] = v[0]["value"]
                    dicts.append(dc)
                df = pd.DataFrame.from_dict(dicts)
                await cls.createGraphFromDataFrame(
                    cls,
                    graph_name=graph_name,
                    src=df,
                    existing_node_ids=existing_node_ids,
                    first_chunk=first_chunk,
                    start_v_label=out_v_label,
                    start_id=id_map[out_v_label],
                    start_props=[
                        k
                        for k in records[1]["properties"].keys()
                        if k != "pk" and k != id_map[out_v_label]
                    ],
                    edge_type=edge_type,
                    end_v_label=in_v_label,
                    end_id=id_map[in_v_label],
                    end_props=[
                        k
                        for k in records[0]["properties"].keys()
                        if k != "pk" and k != id_map[in_v_label]
                    ],
                    chunk_size=chunk_size,
                    direct_loading=direct_loading,
                    drop_graph=drop_graph,
                    use_copy=use_copy,
                )
                first_chunk = False

        g.close()
